# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from json import dumps, loads

# Disable logging for a cleaner testing output
import logging
import os
from unittest import mock

from helpers import AppriseURLTester
import pytest
import requests

from apprise import Apprise, AppriseAttachment, NotifyType
from apprise.plugins.sendpulse import NotifySendPulse

logging.disable(logging.CRITICAL)

SENDPULSE_GOOD_RESPONSE = dumps(
    {
        "access_token": "abc123",
        "expires_in": 3600,
    }
)

SENDPULSE_BAD_RESPONSE = "{"

# Attachment Directory
TEST_VAR_DIR = os.path.join(os.path.dirname(__file__), "var")

# Our Testing URLs
apprise_url_tests = (
    (
        "sendpulse://",
        {
            "instance": TypeError,
        },
    ),
    (
        "sendpulse://:@/",
        {
            "instance": TypeError,
        },
    ),
    (
        "sendpulse://abcd",
        {
            # invalid from email
            "instance": TypeError,
        },
    ),
    (
        "sendpulse://abcd@host.com",
        {
            # Just an Email specified, no client_id or client_secret
            "instance": TypeError,
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs/?template=invalid",
        {
            # Invalid template
            "instance": TypeError,
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs1/?template=123",
        {
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs1/",
        {
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs1/?format=text",
        {
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs1/?format=html",
        {
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://chris@example.com/client_id/cs1/?from=Chris",
        {
            # Set name only
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
            "force_debug": True,
        },
    ),
    (
        "sendpulse://?id=ci&secret=cs&user=chris@example.com",
        {
            # Set login through user= only
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://?id=ci&secret=cs&user=Chris<chris@example.com>",
        {
            # Set login through user= only
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://?id=ci&secret=cs&user=chris",
        {
            # Set login through user= only - invaild email
            "instance": TypeError,
        },
    ),
    (
        "sendpulse://example.com/client_id/cs1/?user=chris",
        {
            # Set user as a name only
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://client_id/cs1/?user=chris@example.ca",
        {
            # Set user as email
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://client_id/cs1/?from=Chris<chris@example.com>",
        {
            # set full email with name
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://?from=Chris<chris@example.com>&id=ci&secret=cs",
        {
            # leverage all get params from URL
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs1a/",
        {
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_BAD_RESPONSE,
            # Notify will fail because auth failed
            "response": False,
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs2/?bcc=l2g@nuxref.com",
        {
            # A good email with Blind Carbon Copy
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs2/?bcc=invalid",
        {
            # A good email with Blind Carbon Copy
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs3/?cc=l2g@nuxref.com",
        {
            # A good email with Carbon Copy
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs4/?cc=Chris<l2g@nuxref.com>",
        {
            # A good email with Carbon Copy
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs5/?cc=invalid",
        {
            # A good email with Carbon Copy
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs6/?to=invalid",
        {
            # an invalid to email
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs7/chris@example.com",
        {
            # An email with a designated to email
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs8/"
        "?to=Chris<chris@example.com>",
        {
            # An email with a full name in in To field
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs9/"
        "chris@example.com/chris2@example.com/Test<test@test.com>",
        {
            # Several emails to notify
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs10/"
        "?cc=Chris<chris@example.com>",
        {
            # An email with a full name in cc
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs11/?cc=chris@example.com",
        {
            # An email with a full name in cc
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs12/"
        "?bcc=Chris<chris@example.com>",
        {
            # An email with a full name in bcc
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs13/?bcc=chris@example.com",
        {
            # An email with a full name in bcc
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs14/"
        "?to=Chris<chris@example.com>",
        {
            # An email with a full name in bcc
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs15/?to=chris@example.com",
        {
            # An email with a full name in bcc
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs16/"
        "?template=1234&+sub=value&+sub2=value2",
        {
            # A good email with a template + substitutions
            "instance": NotifySendPulse,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": "sendpulse://user@example.com/c...d/c...6/",
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs17/",
        {
            "instance": NotifySendPulse,
            # force a failure
            "response": False,
            "requests_response_code": requests.codes.internal_server_error,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs18/",
        {
            "instance": NotifySendPulse,
            # throw a bizarre code forcing us to fail to look it up
            "response": False,
            "requests_response_code": 999,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
    (
        "sendpulse://user@example.com/client_id/cs19/",
        {
            "instance": NotifySendPulse,
            # Throws connection and transfer exceptions when this
            # flag is set and tests that we gracefully handle them
            "test_requests_exceptions": True,
            "requests_response_text": SENDPULSE_GOOD_RESPONSE,
        },
    ),
)


def test_plugin_sendpulse_urls():
    """
    NotifySendPulse() Apprise URLs

    """

    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()


@mock.patch("requests.post")
def test_plugin_sendpulse_edge_cases(mock_post):
    """
    NotifySendPulse() Edge Cases
    """
    request = mock.Mock()
    request.status_code = requests.codes.ok
    request.content = SENDPULSE_GOOD_RESPONSE

    # Prepare Mock
    mock_post.return_value = request

    obj = Apprise.instantiate(
        "sendpulse://user@example.com/ci/cs/Test<test@example.com>"
    )

    assert (
        obj.notify(body="body", title="title", notify_type=NotifyType.INFO)
        is True
    )

    # Test our call count
    assert mock_post.call_count == 2

    # Authentication
    assert (
        mock_post.call_args_list[0][0][0]
        == "https://api.sendpulse.com/oauth/access_token"
    )

    payload = loads(mock_post.call_args_list[0][1]["data"])
    assert payload == {
        "grant_type": "client_credentials",
        "client_id": "ci",
        "client_secret": "cs",
    }

    assert (
        mock_post.call_args_list[1][0][0]
        == "https://api.sendpulse.com/smtp/emails"
    )
    payload = loads(mock_post.call_args_list[1][1]["data"])

    assert payload == {
        "email": {
            "from": {"email": "user@example.com", "name": "Apprise"},
            "to": [{"email": "test@example.com", "name": "Test"}],
            "subject": "title",
            "text": "body",
            "html": "Ym9keQ==",
        }
    }

    mock_post.reset_mock()

    obj = Apprise.instantiate("sendpulse://user@example.com/ci/cs/?from=John")

    assert (
        obj.notify(body="body", title="title", notify_type=NotifyType.INFO)
        is True
    )

    # Test our call count
    assert mock_post.call_count == 2

    # Authentication
    assert (
        mock_post.call_args_list[0][0][0]
        == "https://api.sendpulse.com/oauth/access_token"
    )

    payload = loads(mock_post.call_args_list[0][1]["data"])
    assert payload == {
        "grant_type": "client_credentials",
        "client_id": "ci",
        "client_secret": "cs",
    }

    assert (
        mock_post.call_args_list[1][0][0]
        == "https://api.sendpulse.com/smtp/emails"
    )
    payload = loads(mock_post.call_args_list[1][1]["data"])

    assert payload == {
        "email": {
            "from": {"email": "user@example.com", "name": "John"},
            "to": [{"email": "user@example.com", "name": "John"}],
            "subject": "title",
            "text": "body",
            "html": "Ym9keQ==",
        }
    }

    mock_post.reset_mock()

    # Second call no longer needs to authenticate
    assert (
        obj.notify(body="body", title="title", notify_type=NotifyType.INFO)
        is True
    )

    assert mock_post.call_count == 1

    assert (
        mock_post.call_args_list[0][0][0]
        == "https://api.sendpulse.com/smtp/emails"
    )

    # force an exception
    mock_post.side_effect = requests.RequestException
    assert (
        obj.notify(body="body", title="title", notify_type=NotifyType.INFO)
        is False
    )

    # Set an invalid return code
    mock_post.side_effect = None
    request.status_code = 403
    assert (
        obj.notify(body="body", title="title", notify_type=NotifyType.INFO)
        is False
    )

    # Test re-authentication
    mock_post.reset_mock()
    request = mock.Mock()
    obj = Apprise.instantiate("sendpulse://usr2@example.com/ci/cs/?from=Retry")

    class sendpulse:
        def __init__(self):
            # 200 login okay
            # 401 on retrival
            # recursive re-attempt to login returns 200
            # fetch after works
            self._side_effect = iter(
                [
                    requests.codes.ok,
                    requests.codes.unauthorized,
                    requests.codes.ok,
                    requests.codes.ok,
                ]
            )

        @property
        def status_code(self):
            return next(self._side_effect)

        @property
        def content(self):
            return SENDPULSE_GOOD_RESPONSE

    mock_post.return_value = sendpulse()

    assert (
        obj.notify(body="body", title="title", notify_type=NotifyType.INFO)
        is True
    )

    assert mock_post.call_count == 4
    # Authentication
    assert (
        mock_post.call_args_list[0][0][0]
        == "https://api.sendpulse.com/oauth/access_token"
    )
    # 401 received
    assert (
        mock_post.call_args_list[1][0][0]
        == "https://api.sendpulse.com/smtp/emails"
    )
    # Re-authenticate
    assert (
        mock_post.call_args_list[2][0][0]
        == "https://api.sendpulse.com/oauth/access_token"
    )
    # Try again
    assert (
        mock_post.call_args_list[3][0][0]
        == "https://api.sendpulse.com/smtp/emails"
    )

    # Test re-authentication  (no recursive loops)
    mock_post.reset_mock()
    request = mock.Mock()
    obj = Apprise.instantiate("sendpulse://usr2@example.com/ci/cs/?from=Retry")

    class sendpulse:
        def __init__(self):
            # oauth always returns okay but notify returns 401
            # recursive re-attempt only once
            self._side_effect = iter(
                [
                    requests.codes.ok,
                    requests.codes.unauthorized,
                    requests.codes.ok,
                    requests.codes.unauthorized,
                    requests.codes.ok,
                    requests.codes.unauthorized,
                    requests.codes.ok,
                    requests.codes.unauthorized,
                    requests.codes.ok,
                    requests.codes.unauthorized,
                    requests.codes.ok,
                    requests.codes.unauthorized,
                    requests.codes.ok,
                    requests.codes.unauthorized,
                    requests.codes.ok,
                    requests.codes.unauthorized,
                ]
            )

        @property
        def status_code(self):
            return next(self._side_effect)

        @property
        def content(self):
            return SENDPULSE_GOOD_RESPONSE

    mock_post.return_value = sendpulse()

    assert (
        obj.notify(body="body", title="title", notify_type=NotifyType.INFO)
        is False
    )

    assert mock_post.call_count == 4
    # Authentication
    assert (
        mock_post.call_args_list[0][0][0]
        == "https://api.sendpulse.com/oauth/access_token"
    )
    # 401 received
    assert (
        mock_post.call_args_list[1][0][0]
        == "https://api.sendpulse.com/smtp/emails"
    )
    # Re-authenticate
    assert (
        mock_post.call_args_list[2][0][0]
        == "https://api.sendpulse.com/oauth/access_token"
    )
    # Last failed attempt
    assert (
        mock_post.call_args_list[3][0][0]
        == "https://api.sendpulse.com/smtp/emails"
    )

    mock_post.side_effect = None
    request = mock.Mock()
    request.status_code = requests.codes.ok
    request.content = SENDPULSE_GOOD_RESPONSE
    mock_post.return_value = request
    for expires_in in (None, -1, "garbage", 3600, 300000):
        request.content = dumps(
            {
                "access_token": "abc123",
                "expires_in": expires_in,
            }
        )

        # Instantiate our object
        obj = Apprise.instantiate("sendpulse://user@example.com/ci/cs/")

        # Test variations of responses
        obj.notify(body="body", title="title", notify_type=NotifyType.INFO)

        # expires_in is missing
        request.content = dumps(
            {
                "access_token": "abc123",
            }
        )

        # Instantiate our object
        obj = Apprise.instantiate("sendpulse://user@example.com/ci/cs/")
        assert (
            obj.notify(body="body", title="title", notify_type=NotifyType.INFO)
            is True
        )


def test_plugin_sendpulse_fail_cases():
    """
    NotifySendPulse() Fail Cases

    """

    # no client_id
    with pytest.raises(TypeError):
        NotifySendPulse(
            client_id="abcd", client_secret=None, from_addr="user@example.com"
        )

    with pytest.raises(TypeError):
        NotifySendPulse(
            client_id=None,
            client_secret="abcd123",
            from_addr="user@example.com",
        )

    # invalid from email
    with pytest.raises(TypeError):
        NotifySendPulse(
            client_id="abcd", client_secret="abcd456", from_addr="!invalid"
        )

    # no email
    with pytest.raises(TypeError):
        NotifySendPulse(
            client_id="abcd", client_secret="abcd789", from_addr=None
        )

    # Invalid To email address
    NotifySendPulse(
        client_id="abcd",
        client_secret="abcd321",
        from_addr="user@example.com",
        targets="!invalid",
    )

    # Test invalid bcc/cc entries mixed with good ones
    assert isinstance(
        NotifySendPulse(
            client_id="abcd",
            client_secret="abcd654",
            from_addr="l2g@example.com",
            bcc=("abc@def.com", "!invalid"),
            cc=("abc@test.org", "!invalid"),
        ),
        NotifySendPulse,
    )


@mock.patch("requests.post")
def test_plugin_sendpulse_attachments(mock_post):
    """
    NotifySendPulse() Attachments

    """

    request = mock.Mock()
    request.status_code = requests.codes.ok
    request.content = SENDPULSE_GOOD_RESPONSE

    # Prepare Mock
    mock_post.return_value = request

    path = os.path.join(TEST_VAR_DIR, "apprise-test.gif")
    attach = AppriseAttachment(path)
    obj = Apprise.instantiate("sendpulse://user@example.com/aaaa/bbbb")
    assert isinstance(obj, NotifySendPulse)
    assert (
        obj.notify(
            body="body",
            title="title",
            notify_type=NotifyType.INFO,
            attach=attach,
        )
        is True
    )

    mock_post.reset_mock()

    # Try again in a use case where we can't access the file
    with mock.patch("os.path.isfile", return_value=False):
        assert (
            obj.notify(
                body="body",
                title="title",
                notify_type=NotifyType.INFO,
                attach=attach,
            )
            is False
        )

    # Try again in a use case where we can't access the file
    with mock.patch("builtins.open", side_effect=OSError):
        assert (
            obj.notify(
                body="body",
                title="title",
                notify_type=NotifyType.INFO,
                attach=attach,
            )
            is False
        )



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Disable logging for a cleaner testing output
import logging
from unittest import mock

from helpers import AppriseURLTester
import requests

from apprise.plugins.synology import NotifySynology

logging.disable(logging.CRITICAL)

# Our Testing URLs
apprise_url_tests = (
    (
        "synology://:@/",
        {
            "instance": None,
        },
    ),
    (
        "synology://",
        {
            "instance": None,
        },
    ),
    (
        "synologys://",
        {
            "instance": None,
        },
    ),
    (
        "synology://localhost/token",
        {
            "instance": NotifySynology,
        },
    ),
    (
        "synology://localhost/token?file_url=http://reddit.com/test.jpg",
        {
            "instance": NotifySynology,
        },
    ),
    (
        "synology://user:pass@localhost/token",
        {
            "instance": NotifySynology,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": "synology://user:****@localhost/t...n",
        },
    ),
    # No token was specified
    (
        "synology://user@localhost",
        {
            "instance": TypeError,
        },
    ),
    (
        "synology://user@localhost/token",
        {
            "instance": NotifySynology,
        },
    ),
    # Continue testing other cases
    (
        "synology://localhost:8080/token",
        {
            "instance": NotifySynology,
        },
    ),
    (
        "synology://user:pass@localhost:8080/token",
        {
            "instance": NotifySynology,
        },
    ),
    (
        "synologys://localhost/token",
        {
            "instance": NotifySynology,
        },
    ),
    (
        "synologys://localhost/?token=mytoken",
        {
            "instance": NotifySynology,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": "synologys://localhost/m...n",
        },
    ),
    (
        "synologys://user:pass@localhost/token",
        {
            "instance": NotifySynology,
        },
    ),
    (
        "synologys://localhost:8080/token/path/",
        {
            "instance": NotifySynology,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": "synologys://localhost:8080/t...n/path/",
        },
    ),
    (
        "synologys://user:password@localhost:8080/token",
        {
            "instance": NotifySynology,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": "synologys://user:****@localhost:8080/t...n",
        },
    ),
    # Test our Headers
    (
        "synology://localhost:8080/path?+HeaderKey=HeaderValue",
        {
            "instance": NotifySynology,
        },
    ),
    (
        "synology://user:pass@localhost:8081/token",
        {
            "instance": NotifySynology,
            # force a failure
            "response": False,
            "requests_response_code": requests.codes.internal_server_error,
        },
    ),
    (
        "synology://user:pass@localhost:8082/token",
        {
            "instance": NotifySynology,
            # throw a bizarre code forcing us to fail to look it up
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        "synology://user:pass@localhost:8083/token",
        {
            "instance": NotifySynology,
            # Throws a series of i/o exceptions with this flag
            # is set and tests that we gracefully handle them
            "test_requests_exceptions": True,
        },
    ),
)


def test_plugin_custom_synology_urls():
    """NotifySynology() Apprise URLs."""

    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()


@mock.patch("requests.post")
def test_plugin_synology_edge_cases(mock_post):
    """NotifySynology() Edge Cases."""

    # Prepare our response
    response = requests.Request()
    response.status_code = requests.codes.ok

    # Prepare Mock
    mock_post.return_value = response

    # This string also tests that type is set to nothing
    results = NotifySynology.parse_url(
        "synology://user:pass@localhost:8080/token"
    )

    assert isinstance(results, dict)
    assert results["user"] == "user"
    assert results["password"] == "pass"
    assert results["port"] == 8080
    assert results["host"] == "localhost"
    assert results["fullpath"] == ""
    assert results["path"] == "/"
    assert results["query"] == "token"
    assert results["schema"] == "synology"
    assert results["url"] == "synology://user:pass@localhost:8080/token"

    instance = NotifySynology(**results)
    assert isinstance(instance, NotifySynology)

    response = instance.send(title="title", body="body")
    assert response is True
    assert mock_post.call_count == 1

    details = mock_post.call_args_list[0]
    assert details[0][0] == "http://localhost:8080/webapi/entry.cgi"

    assert details[1]["data"].startswith("payload=")



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging
import sys
from unittest import mock

from helpers import AppriseURLTester
import pytest

from apprise import Apprise, NotifyType
from apprise.plugins.blink1 import (
    BLINK1_DEFAULT_DURATION_MS,
    BLINK1_DEFAULT_FADE_MS,
    BLINK1_MAX_DURATION_MS,
    BLINK1_MAX_FADE_MS,
    BLINK1_REPORT_SIZE,
    Blink1LED,
    NotifyBlink1,
    _blink1_fade_buf,
)

logging.disable(logging.CRITICAL)

# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def _mk_device(report_size=BLINK1_REPORT_SIZE, open_exc=None):
    """Return a mock hidapi device that succeeds by default."""
    dev = mock.Mock()
    dev.send_feature_report.return_value = report_size
    if open_exc is not None:
        dev.open.side_effect = open_exc
    return dev


# ---------------------------------------------------------------------------
# URL matrix used by AppriseURLTester
# ---------------------------------------------------------------------------

apprise_url_tests = (
    # First available device, all defaults
    (
        "blink1://",
        {
            "instance": NotifyBlink1,
        },
    ),
    # Explicit underscore placeholder for first device
    (
        "blink1://_/",
        {
            "instance": NotifyBlink1,
        },
    ),
    # Specific device by serial number
    (
        "blink1://ABCD1234/",
        {
            "instance": NotifyBlink1,
        },
    ),
    # Minimum allowed parameter values
    (
        "blink1://?duration=0&fade=0&ledn=0",
        {
            "instance": NotifyBlink1,
        },
    ),
    # LED 1 (first LED only)
    (
        "blink1://?ledn=1",
        {
            "instance": NotifyBlink1,
        },
    ),
    # LED 2 (second LED only)
    (
        "blink1://?ledn=2",
        {
            "instance": NotifyBlink1,
        },
    ),
    # Maximum allowed duration
    (
        f"blink1://?duration={BLINK1_MAX_DURATION_MS}",
        {
            "instance": NotifyBlink1,
        },
    ),
    # Maximum allowed fade
    (
        f"blink1://?fade={BLINK1_MAX_FADE_MS}",
        {
            "instance": NotifyBlink1,
        },
    ),
    # Non-numeric duration -> TypeError
    (
        "blink1://?duration=abc",
        {
            "instance": TypeError,
        },
    ),
    # Negative duration -> TypeError
    (
        "blink1://?duration=-1",
        {
            "instance": TypeError,
        },
    ),
    # Duration exceeds maximum -> TypeError
    (
        f"blink1://?duration={BLINK1_MAX_DURATION_MS + 1}",
        {
            "instance": TypeError,
        },
    ),
    # Non-numeric fade -> TypeError
    (
        "blink1://?fade=abc",
        {
            "instance": TypeError,
        },
    ),
    # Negative fade -> TypeError
    (
        "blink1://?fade=-1",
        {
            "instance": TypeError,
        },
    ),
    # Fade exceeds maximum -> TypeError
    (
        f"blink1://?fade={BLINK1_MAX_FADE_MS + 1}",
        {
            "instance": TypeError,
        },
    ),
)


# ---------------------------------------------------------------------------
# Import-error path (run only when hidapi is NOT installed)
# ---------------------------------------------------------------------------


@pytest.mark.skipif(
    "hid" in sys.modules,
    reason="Requires that hidapi NOT be installed",
)
def test_plugin_blink1_import_error():
    """NotifyBlink1() returns None when hidapi is absent."""

    obj = Apprise.instantiate("blink1://")
    assert obj is None


# ---------------------------------------------------------------------------
# All remaining tests require hidapi to be installed
# ---------------------------------------------------------------------------


def test_plugin_blink1_urls():
    """NotifyBlink1() Apprise URL matrix."""

    pytest.importorskip("hid")

    with (
        mock.patch("apprise.plugins.blink1.blink1_hid") as mock_hid,
        mock.patch("apprise.plugins.blink1.time") as mock_time,
    ):
        mock_dev = _mk_device()
        mock_hid.device.return_value = mock_dev
        mock_time.sleep.return_value = None

        AppriseURLTester(tests=apprise_url_tests).run_all()


def test_plugin_blink1_device_open_failure():
    """NotifyBlink1() returns False when the USB device cannot be opened."""

    pytest.importorskip("hid")

    with mock.patch("apprise.plugins.blink1.blink1_hid") as mock_hid:
        mock_dev = _mk_device(open_exc=OSError("no device"))
        mock_hid.device.return_value = mock_dev

        obj = Apprise.instantiate("blink1://ABCD1234/")
        assert obj is not None
        assert obj.notify(body="test") is False


def test_plugin_blink1_device_open_failure_no_serial():
    """NotifyBlink1() failure log omits serial when none is configured."""

    pytest.importorskip("hid")

    with mock.patch("apprise.plugins.blink1.blink1_hid") as mock_hid:
        mock_dev = _mk_device(open_exc=OSError("no device"))
        mock_hid.device.return_value = mock_dev

        obj = Apprise.instantiate("blink1://")
        assert obj is not None
        assert obj.notify(body="test") is False


def test_plugin_blink1_send_feature_report_exception():
    """NotifyBlink1() returns False and closes device on HID write error."""

    pytest.importorskip("hid")

    with (
        mock.patch("apprise.plugins.blink1.blink1_hid") as mock_hid,
        mock.patch("apprise.plugins.blink1.time") as mock_time,
    ):
        mock_dev = mock.Mock()
        mock_dev.send_feature_report.side_effect = OSError("USB error")
        mock_hid.device.return_value = mock_dev
        mock_time.sleep.return_value = None

        obj = Apprise.instantiate("blink1://")
        assert obj is not None
        assert obj.notify(body="test") is False

        # close() must always be called
        assert mock_dev.close.called


def test_plugin_blink1_short_report_fails():
    """NotifyBlink1() returns False when the HID report is truncated."""

    pytest.importorskip("hid")

    with (
        mock.patch("apprise.plugins.blink1.blink1_hid") as mock_hid,
        mock.patch("apprise.plugins.blink1.time") as mock_time,
    ):
        mock_dev = _mk_device(report_size=BLINK1_REPORT_SIZE - 1)
        mock_hid.device.return_value = mock_dev
        mock_time.sleep.return_value = None

        obj = Apprise.instantiate("blink1://")
        assert obj is not None
        assert obj.notify(body="test") is False


def test_plugin_blink1_notify_types():
    """NotifyBlink1() sends one fade per notification type."""

    pytest.importorskip("hid")

    with (
        mock.patch("apprise.plugins.blink1.blink1_hid") as mock_hid,
        mock.patch("apprise.plugins.blink1.time") as mock_time,
    ):
        mock_dev = _mk_device()
        mock_hid.device.return_value = mock_dev
        mock_time.sleep.return_value = None

        obj = Apprise.instantiate("blink1://ABCD1234/?fade=100&duration=50")
        assert obj is not None

        for notify_type in (
            NotifyType.INFO,
            NotifyType.SUCCESS,
            NotifyType.WARNING,
            NotifyType.FAILURE,
        ):
            assert obj.notify(body="test", notify_type=notify_type) is True

        # Each notification sends two reports (color + off)
        assert mock_dev.send_feature_report.call_count == 8
        assert mock_dev.close.call_count == 4


def test_plugin_blink1_fade_buf_layout():
    """_blink1_fade_buf() constructs the correct 9-byte HID payload."""

    buf = _blink1_fade_buf(255, 128, 0, 1000, ledn=1)

    assert len(buf) == 9
    assert buf[0] == 0x01  # REPORT_ID
    assert buf[1] == ord("c")  # command byte
    assert buf[2] == 255  # red
    assert buf[3] == 128  # green
    assert buf[4] == 0  # blue
    # fade_time = 1000 // 10 = 100 = 0x0064
    assert buf[5] == 0x00  # th
    assert buf[6] == 0x64  # tl
    assert buf[7] == 1  # ledn
    assert buf[8] == 0x00  # padding


def test_plugin_blink1_fade_rgb_args():
    """NotifyBlink1() passes correct ledn and RGB values to the device."""

    pytest.importorskip("hid")

    with (
        mock.patch("apprise.plugins.blink1.blink1_hid") as mock_hid,
        mock.patch("apprise.plugins.blink1.time") as mock_time,
    ):
        mock_dev = _mk_device()
        mock_hid.device.return_value = mock_dev
        mock_time.sleep.return_value = None

        obj = Apprise.instantiate("blink1://?fade=200&duration=100&ledn=2")
        assert obj is not None
        assert obj.notify(body="test", notify_type=NotifyType.INFO) is True

        first_call_buf = mock_dev.send_feature_report.call_args_list[0][0][0]
        assert first_call_buf[1] == ord("c")  # command byte
        assert first_call_buf[7] == 2  # ledn
        # fade_time = 200 // 10 = 20 = 0x0014
        assert first_call_buf[5] == 0x00
        assert first_call_buf[6] == 0x14


def test_plugin_blink1_off_sent_after_duration():
    """NotifyBlink1() sends an off command after sleeping for fade+duration."""

    pytest.importorskip("hid")

    with (
        mock.patch("apprise.plugins.blink1.blink1_hid") as mock_hid,
        mock.patch("apprise.plugins.blink1.time") as mock_time,
    ):
        mock_dev = _mk_device()
        mock_hid.device.return_value = mock_dev
        mock_time.sleep.return_value = None

        obj = Apprise.instantiate("blink1://?fade=500&duration=2000")
        assert obj is not None
        assert obj.notify(body="test") is True

        mock_time.sleep.assert_called_once_with(2.5)

        # Second report should be the "off" command (all zeros for RGB)
        off_buf = mock_dev.send_feature_report.call_args_list[1][0][0]
        assert off_buf[2] == 0  # red
        assert off_buf[3] == 0  # green
        assert off_buf[4] == 0  # blue


def test_plugin_blink1_off_command_failure_returns_false():
    """NotifyBlink1() returns False when the off HID write fails."""

    pytest.importorskip("hid")

    with (
        mock.patch("apprise.plugins.blink1.blink1_hid") as mock_hid,
        mock.patch("apprise.plugins.blink1.time") as mock_time,
    ):
        mock_dev = mock.Mock()
        # First call (color fade) succeeds; second call (off) fails.
        mock_dev.send_feature_report.side_effect = [
            BLINK1_REPORT_SIZE,
            BLINK1_REPORT_SIZE - 1,
        ]
        mock_hid.device.return_value = mock_dev
        mock_time.sleep.return_value = None

        obj = Apprise.instantiate("blink1://")
        assert obj is not None
        assert obj.notify(body="test") is False

        # close() must still be called
        assert mock_dev.close.called


def test_plugin_blink1_url_round_trip():
    """NotifyBlink1() URL reconstructs an equivalent object."""

    obj = NotifyBlink1(serial="ABCD1234", duration=3000, fade=500, ledn=1)
    rebuilt = NotifyBlink1(**NotifyBlink1.parse_url(obj.url()))

    assert rebuilt.serial == "ABCD1234"
    assert rebuilt.duration == 3000
    assert rebuilt.fade == 500
    assert rebuilt.ledn == 1


def test_plugin_blink1_url_round_trip_no_serial():
    """NotifyBlink1() URL round-trips correctly without a serial."""

    obj = NotifyBlink1()
    rebuilt = NotifyBlink1(**NotifyBlink1.parse_url(obj.url()))

    assert rebuilt.serial is None
    assert rebuilt.duration == BLINK1_DEFAULT_DURATION_MS
    assert rebuilt.fade == BLINK1_DEFAULT_FADE_MS
    assert rebuilt.ledn == Blink1LED.ALL


def test_plugin_blink1_url_identifier():
    """NotifyBlink1() url_identifier distinguishes different connections."""

    obj_a = NotifyBlink1()
    obj_b = NotifyBlink1(serial="ABCD1234")
    obj_c = NotifyBlink1(ledn=1)

    assert obj_a.url_identifier != obj_b.url_identifier
    assert obj_a.url_identifier != obj_c.url_identifier
    assert obj_b.url_identifier != obj_c.url_identifier


def test_plugin_blink1_invalid_ledn_defaults():
    """NotifyBlink1() silently falls back to ALL for an unrecognised ledn."""

    obj = NotifyBlink1(ledn="99")
    assert obj.ledn == Blink1LED.ALL


def test_plugin_blink1_runtime_deps():
    """NotifyBlink1.runtime_deps() returns the importable hid module name."""

    assert NotifyBlink1.runtime_deps() == ("hid",)


def test_plugin_blink1_parse_url_underscore():
    """NotifyBlink1() treats underscore host as first-device placeholder."""

    result = NotifyBlink1.parse_url("blink1://_/")
    assert result is not None
    assert result.get("serial") is None


def test_plugin_blink1_parse_url_serial():
    """NotifyBlink1() extracts serial number from URL host field."""

    result = NotifyBlink1.parse_url("blink1://DEADBEEF/")
    assert result is not None
    assert result["serial"] == "DEADBEEF"



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
from apprise import NotifyFormat
from apprise.utils.format import html_adjust, markdown_adjust, smart_split


def test_smart_split_prefers_newlines_over_spaces_and_punctuation():
    """
    Newlines should win even if there are spaces and punctuation before the
    limit.
    """
    text = "line1\nline2 line3. line4"
    # Long enough to include the newline and some of the next line
    limit = 12

    chunks = smart_split(text, limit, body_format=NotifyFormat.TEXT)

    # First chunk should end immediately after the newline
    assert chunks[0] == "line1\n"
    # Nothing lost
    assert "".join(chunks) == text


def test_smart_split_prefers_spaces_over_hard_split():
    """
    When there are no newlines, split on the last space/tab before falling back
    to a hard character-limit split.
    """
    text = "word1 word2 word3"
    # Force a split between word2 and word3
    limit = 12  # "word1 word2 " is 12 characters

    chunks = smart_split(text, limit, body_format=NotifyFormat.TEXT)

    assert chunks == ["word1 word2 ", "word3"]
    assert "".join(chunks) == text


def test_smart_split_can_split_after_punctuation_plus_whitespace():
    """
    Exercise the punctuation+whitespace pattern. In practice this collapses
    to the same split point as the last space, but we verify the behaviour.
    """
    text = "Hello world. Again"
    # Force the split around ". "

    # "Hello world. " is 13 characters
    limit = 13

    chunks = smart_split(text, limit, body_format=NotifyFormat.TEXT)

    # First chunk should end at the space after the period
    assert chunks[0] == "Hello world. "
    assert chunks[1] == "Again"
    assert "".join(chunks) == text


def test_smart_split_avoids_splitting_inside_html_entity() -> None:
    """
    In HTML mode we must not end a chunk in the middle of '&...;'.

    We do NOT assert exact chunk values. Instead we assert:
    - TEXT mode can split inside the entity.
    - HTML mode never has a chunk that contains '&' without a matching ';'
      after it in the same chunk.
    """
    text = "1234&nbsp;5678"
    limit = 8  # without adjustment, we would cut inside '&nbsp;'

    # Plain text mode: allowed to split anywhere
    chunks_text = smart_split(text, limit, body_format=NotifyFormat.TEXT)
    assert "".join(chunks_text) == text

    # Sanity: in TEXT mode we *do* split inside the entity
    assert any(
        "&" in chunk and ";" not in chunk[chunk.find("&") :]
        for chunk in chunks_text
    )

    # HTML mode: entity-aware
    chunks_html = smart_split(text, limit, body_format=NotifyFormat.HTML)
    assert "".join(chunks_html) == text

    # If a chunk contains '&', it must also contain the terminating ';'
    # for that entity within the same chunk.
    for chunk in chunks_html:
        idx = chunk.find("&")
        if idx == -1:
            continue
        semi = chunk.find(";", idx + 1)
        assert semi != -1, f"Chunk ends inside HTML entity: {chunk!r}"


def test_smart_split_avoids_splitting_inside_markdown_link() -> None:
    """
    In MARKDOWN mode, do not split inside [text](url).

    We only require that the full [link](...) lies in a single chunk and that
    any '[' appearing in a chunk has a matching ')' in that same chunk.
    """
    link = "[link](https://example.com)"
    text = "AAAA" + link
    limit = len(link)

    chunks = smart_split(text, limit, body_format=NotifyFormat.MARKDOWN)
    assert "".join(chunks) == text

    # Entire link must be contained in one chunk
    assert any(link in chunk for chunk in chunks)

    # If a chunk has '[', it must also contain its closing ')'
    for chunk in chunks:
        idx = chunk.find("[")
        if idx == -1:
            continue
        semi = chunk.find(")", idx + 1)
        assert semi != -1, f"Markdown link was split inside chunk: {chunk!r}"


def test_smart_split_avoids_splitting_inside_markdown_image() -> None:
    """
    In MARKDOWN mode, do not split inside the [alt](url) of an image.

    The implementation currently splits "AAAA![alt](...)" as:
      - "AAAA!"
      - "[alt](...)"
    which is acceptable, as the [alt](url) part is kept intact.
    """
    image = "![alt](https://example.com/image.png)"
    text = "AAAA" + image
    limit = len(image)

    chunks = smart_split(text, limit, body_format=NotifyFormat.MARKDOWN)
    assert "".join(chunks) == text

    inner = "[alt](https://example.com/image.png)"

    # The [alt](...) portion must appear fully within a single chunk
    assert any(inner in chunk for chunk in chunks)

    # As with links, any '[' in a chunk must have its matching ')' within
    # the same chunk so we never split inside the [alt](url) part.
    for chunk in chunks:
        idx = chunk.find("[")
        if idx == -1:
            continue
        semi = chunk.find(")", idx + 1)
        assert semi != -1, f"Markdown image was split inside chunk: {chunk!r}"


def test_smart_split_empty_and_none_input() -> None:
    """
    Empty / None input should be returned as a single-element list unchanged.
    """
    assert smart_split("", 10, body_format=NotifyFormat.TEXT) == [""]
    assert smart_split("", 0, body_format=NotifyFormat.TEXT) == [""]
    assert smart_split("content", 0, body_format=NotifyFormat.TEXT) == [""]
    # None short-circuits before len() is called
    assert smart_split(None, 10, body_format=NotifyFormat.TEXT) == [""]


def test_smart_split_html_entity_exact_boundary() -> None:
    """
    Splitting exactly at an HTML entity boundary should not shift the split
    point (no need to "fix up" a perfectly aligned boundary).
    """
    text = "AAAA&nbsp;BBBB"
    limit = len("AAAA&nbsp;")  # split exactly after the entity

    chunks = smart_split(text, limit, body_format=NotifyFormat.HTML)

    # We expect the entity to remain whole in the first chunk
    assert chunks == ["AAAA&nbsp;", "BBBB"]
    assert "".join(chunks) == text


def test_smart_split_markdown_link_exact_boundary() -> None:
    """
    Splitting exactly after a Markdown link should not cause any adjustment.
    """
    link = "[link](https://example.com)"
    tail = " TAIL"
    text = link + tail
    limit = len(link)  # split immediately after ')'

    chunks = smart_split(text, limit, body_format=NotifyFormat.MARKDOWN)

    # First chunk is exactly the link, second is the remainder
    assert chunks[0] == link
    assert "".join(chunks) == text

    # Sanity: the link itself is not split across chunks
    assert any(link in chunk for chunk in chunks)


def test_smart_split_whitespace_priority_with_tabs_and_newlines() -> None:
    """
    Exercise newline vs space/tab priority with a mix of whitespace.
    """
    text = "word1\tword2\nword3"

    # Case 1: window ends just before the newline, so only tab is visible.
    limit_without_newline = text.index("\n")  # position of '\n'
    chunks_no_nl = smart_split(
        text, limit_without_newline, body_format=NotifyFormat.TEXT
    )
    # First chunk should end after the tab, since that is the last space/tab
    assert chunks_no_nl[0] == "word1\t"
    assert "".join(chunks_no_nl) == text

    # Case 2: window includes the newline; newline should win over tab.
    limit_with_newline = text.index("\n") + 1
    chunks_with_nl = smart_split(
        text, limit_with_newline, body_format=NotifyFormat.TEXT
    )
    # First chunk should now end after the newline
    assert chunks_with_nl[0] == "word1\tword2\n"
    assert "".join(chunks_with_nl) == text


def test_smart_split_very_short_limit() -> None:
    """
    Very small limits should still split deterministically without loss.
    """
    text = "ABC"
    chunks = smart_split(text, 1, body_format=NotifyFormat.TEXT)

    # One character per chunk
    assert chunks == ["A", "B", "C"]
    assert "".join(chunks) == text


def test_smart_split_very_long_limit() -> None:
    """
    Very large limits (>= len(text)) should return a single chunk.
    """
    text = "A short message for testing"
    chunks = smart_split(text, 10_000, body_format=NotifyFormat.TEXT)

    assert chunks == [text]
    assert "".join(chunks) == text


def test_html_adjust_guard_paths_and_no_entity() -> None:
    """
    Cover the early-return guard in html_adjust and the path where there is
    no '&' at all in the search window.
    """
    text = "abcdef"

    # split_at <= window_start -> early-return unchanged
    assert html_adjust(text, window_start=2, split_at=2) == 2

    # split_at beyond the end of the text -> early-return unchanged
    assert (
        html_adjust(text, window_start=0, split_at=len(text) + 5)
        == len(text) + 5
    )

    # No '&' in window, nothing to adjust
    assert html_adjust(text, window_start=0, split_at=3) == 3


def test_html_adjust_inside_and_at_boundary_of_entity() -> None:
    """
    Exercise the path where html_adjust moves the split back to '&' when the
    split falls inside an entity, and the path where the split is exactly at
    the entity boundary and should not move.
    """
    text = "1234&nbsp;5678"
    # indexes: 0..3 '1234', 4 '&', 5 'n', 6 'b', 7 's', 8 'p', 9 ';', 10 '5'...

    # Split inside '&nbsp;' (at index 8) -> move back to '&' (index 4)
    assert html_adjust(text, window_start=0, split_at=8) == 4

    # Split exactly after the ';' (index 10) -> already outside entity
    assert html_adjust(text, window_start=0, split_at=10) == 10


def test_markdown_adjust_guard_and_no_construct() -> None:
    """
    Cover the guard in markdown_adjust and the case where there is no
    '[' or '!' in the window.
    """
    text = "plain text"

    # split_at <= window_start -> early-return unchanged
    assert markdown_adjust(text, window_start=4, split_at=4) == 4

    # split_at past the end -> early-return unchanged
    assert (
        markdown_adjust(text, window_start=0, split_at=len(text) + 3)
        == len(text) + 3
    )

    # No markdown constructs -> nothing to adjust
    assert markdown_adjust(text, window_start=0, split_at=5) == 5


def test_markdown_adjust_inside_construct_moves_to_start() -> None:
    """
    Exercise the positive path in markdown_adjust where the split lands
    inside a [text](url) construct and the function moves the split
    back to the start of the construct.
    """
    link = "[link](https://example.com)"
    # Choose a split point inside the URL
    split_at = link.index("(") + 3  # somewhere inside "(https..."
    adjusted = markdown_adjust(link, window_start=0, split_at=split_at)

    # Should move back to the '[' at index 0
    assert adjusted == 0


def test_smart_split_markdown_guard_split_at_start_is_reset() -> None:
    """
    Cover the smart_split guard 'if split_at <= start: split_at = orig_split'.

    We force markdown_adjust to move the split back to the window start,
    then verify smart_split resets to the original split so progress is
    still made and chunks join back to the original text.
    """
    text = "[link](https://example.com)"
    limit = 5  # will cause the first soft split to land inside the link

    chunks = smart_split(text, limit, body_format=NotifyFormat.MARKDOWN)

    # We should never get stuck; all chunks must be non-empty
    assert len(chunks) >= 2
    assert all(chunks)

    # Re-joining all chunks must restore the original text
    assert "".join(chunks) == text


def test_smart_split_uses_punctuation_branch_on_rare_whitespace() -> None:
    """
    When punctuation is followed by rare whitespace (vertical tab / form feed)
    and there are no spaces/tabs/newlines, we should use the punctuation
    + whitespace split branch.
    """
    vt = "\x0b"  # vertical tab
    text = f"Hello.{vt}World"
    # Window includes 'Hello.' and the VT
    limit = len("Hello.") + 1

    chunks = smart_split(text, limit, body_format=NotifyFormat.TEXT)

    assert "".join(chunks) == text
    # We expect the first chunk to end after the rare whitespace
    assert chunks[0] == f"Hello.{vt}"
    assert chunks[1] == "World"



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Disable logging for a cleaner testing output
import logging
from unittest import mock

from helpers import AppriseURLTester
import requests

from apprise import Apprise
from apprise.plugins.pushdeer import NotifyPushDeer

logging.disable(logging.CRITICAL)

# Our Testing URLs
apprise_url_tests = (
    (
        "pushdeer://",
        {
            "instance": TypeError,
        },
    ),
    (
        "pushdeers://",
        {
            "instance": TypeError,
        },
    ),
    (
        "pushdeer://localhost/{}".format("a" * 8),
        {
            "instance": NotifyPushDeer,
            # throw a bizarre code forcing us to fail to look it up
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        "pushdeer://localhost/{}".format("a" * 8),
        {
            "instance": NotifyPushDeer,
            # Throws a series of i/o exceptions with this flag
            # is set and tests that we gracefully handle them
            "test_requests_exceptions": True,
        },
    ),
    (
        "pushdeer://localhost:80/{}".format("a" * 8),
        {
            "instance": NotifyPushDeer,
            # throw a bizarre code forcing us to fail to look it up
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        "pushdeer://localhost:80/{}".format("a" * 8),
        {
            "instance": NotifyPushDeer,
            # Throws a series of i/o exceptions with this flag
            # is set and tests that we gracefully handle them
            "test_requests_exceptions": True,
        },
    ),
    (
        "pushdeer://{}".format("a" * 8),
        {
            "instance": NotifyPushDeer,
            # throw a bizarre code forcing us to fail to look it up
            "response": False,
            "requests_response_code": requests.codes.internal_server_error,
        },
    ),
    (
        "pushdeer://{}".format("a" * 8),
        {
            "instance": NotifyPushDeer,
            # Throws a series of i/o exceptions with this flag
            # is set and tests that we gracefully handle them
            "test_requests_exceptions": True,
        },
    ),
)


def test_plugin_pushdeer_urls():
    """NotifyPushDeer() Apprise URLs."""

    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()


@mock.patch("requests.post")
def test_plugin_pushdeer_general(mock_post):
    """NotifyPushDeer() General Checks."""

    response = mock.Mock()
    response.content = ""
    response.status_code = requests.codes.ok

    # Prepare Mock
    mock_post.return_value = response

    # Variation Initializations
    obj = Apprise.instantiate("pushdeer://localhost/pushKey")
    assert isinstance(obj, NotifyPushDeer)
    assert isinstance(obj.url(), str)

    # Send Notification
    assert obj.send(body="test") is True

    assert mock_post.call_count == 1
    assert (
        mock_post.call_args_list[0][0][0]
        == "http://localhost:80/message/push?pushkey=pushKey"
    )



#
# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Disable logging for a cleaner testing output
from json import dumps, loads
import logging
from unittest import mock

from helpers import AppriseURLTester
import requests

from apprise import Apprise
from apprise.common import NotifyFormat
from apprise.plugins.pushplus import (
    PUSHPLUS_CHANNEL_DEFAULT,
    PUSHPLUS_CHANNELS,
    PUSHPLUS_FORMAT_MAP,
    NotifyPushplus,
    PushPlusChannel,
)

logging.disable(logging.CRITICAL)

# A 32-character token used throughout the tests (minimum valid length)
GOOD_TOKEN = "abc123def456ghi789jkl012mno345pq"

# A 64-character token (upper boundary of the allowed length range)
LONG_TOKEN = "a" * 64

# PushPlus always returns HTTP 200; success/failure lives in the JSON body
GOOD_RESPONSE = dumps({"code": 200, "msg": "ok", "data": "msgid"})
BAD_RESPONSE = dumps({"code": 907, "msg": "Token does not exist."})

# Our Testing URLs
apprise_url_tests = (
    # ----------------------------------------------------------------
    # Invalid / missing token cases
    # ----------------------------------------------------------------
    (
        "pushplus://",
        {
            # Empty token must raise
            "instance": TypeError,
        },
    ),
    (
        "pushplus://short",
        {
            # Token too short (fewer than 32 characters)
            "instance": TypeError,
        },
    ),
    (
        "pushplus://invalid!chars00000000000000000000000000000",
        {
            # Token contains characters not allowed by the regex
            "instance": TypeError,
        },
    ),
    # ----------------------------------------------------------------
    # Valid token -- basic personal notification (no targets)
    # ----------------------------------------------------------------
    (
        "pushplus://{}".format(GOOD_TOKEN),
        {
            "instance": NotifyPushplus,
            # Privacy URL must mask the token
            "privacy_url": "pushplus://****/",
            "requests_response_text": GOOD_RESPONSE,
        },
    ),
    # Token supplied as a query parameter instead of the URL host
    (
        "pushplus://?token={}".format(GOOD_TOKEN),
        {
            "instance": NotifyPushplus,
            "privacy_url": "pushplus://****/",
            "requests_response_text": GOOD_RESPONSE,
        },
    ),
    # 64-character (maximum-length) token
    (
        "pushplus://{}".format(LONG_TOKEN),
        {
            "instance": NotifyPushplus,
            "requests_response_text": GOOD_RESPONSE,
        },
    ),
    # ----------------------------------------------------------------
    # Group topic variations
    # ----------------------------------------------------------------
    # Single topic in the URL path (no prefix required)
    (
        "pushplus://{}/mygroup".format(GOOD_TOKEN),
        {
            "instance": NotifyPushplus,
            "privacy_url": "pushplus://****/mygroup/",
            "requests_response_text": GOOD_RESPONSE,
        },
    ),
    # Two topics in the path -- two API calls are made
    (
        "pushplus://{}/group1/group2".format(GOOD_TOKEN),
        {
            "instance": NotifyPushplus,
            "requests_response_text": GOOD_RESPONSE,
        },
    ),
    # Topic supplied via the backward-compatible ?topic= parameter
    (
        "pushplus://{}/?topic=mygroup".format(GOOD_TOKEN),
        {
            "instance": NotifyPushplus,
            "requests_response_text": GOOD_RESPONSE,
        },
    ),
    # Topic supplied via the ?to= convenience alias
    (
        "pushplus://{}/?to=mygroup".format(GOOD_TOKEN),
        {
            "instance": NotifyPushplus,
            "requests_response_text": GOOD_RESPONSE,
        },
    ),
    # Invalid topic characters are silently moved to invalid_targets;
    # the object still instantiates and sends a personal notification
    (
        "pushplus://{}/bad!topic".format(GOOD_TOKEN),
        {
            "instance": NotifyPushplus,
            "requests_response_text": GOOD_RESPONSE,
        },
    ),
    # ----------------------------------------------------------------
    # Delivery channel via ?channel= query parameter
    # ----------------------------------------------------------------
    (
        "pushplus://{}?channel=mail".format(GOOD_TOKEN),
        {
            "instance": NotifyPushplus,
            "requests_response_text": GOOD_RESPONSE,
        },
    ),
    (
        "pushplus://{}?channel=sms".format(GOOD_TOKEN),
        {
            "instance": NotifyPushplus,
            "requests_response_text": GOOD_RESPONSE,
        },
    ),
    (
        "pushplus://{}?channel=cp".format(GOOD_TOKEN),
        {
            "instance": NotifyPushplus,
            "requests_response_text": GOOD_RESPONSE,
        },
    ),
    # wecom is a friendly alias that maps to the "cp" API channel
    (
        "pushplus://{}?channel=wecom".format(GOOD_TOKEN),
        {
            "instance": NotifyPushplus,
            "requests_response_text": GOOD_RESPONSE,
        },
    ),
    # mode= is a synonym for channel=
    (
        "pushplus://{}?mode=mail".format(GOOD_TOKEN),
        {
            "instance": NotifyPushplus,
            "requests_response_text": GOOD_RESPONSE,
        },
    ),
    # ----------------------------------------------------------------
    # Schema alias that auto-sets the delivery channel
    # ----------------------------------------------------------------
    # wecom:// auto-sets channel=cp (WeCom / Enterprise WeChat)
    (
        "wecom://{}".format(GOOD_TOKEN),
        {
            "instance": NotifyPushplus,
            "requests_response_text": GOOD_RESPONSE,
        },
    ),
    # ----------------------------------------------------------------
    # Webhook channel with a named endpoint -- both input forms accepted
    # ----------------------------------------------------------------
    (
        "pushplus://{}?channel=webhook&name=myhook".format(GOOD_TOKEN),
        {
            "instance": NotifyPushplus,
            # url() emits the compact schema://name@token form
            "privacy_url": "pushplus://myhook@",
            "requests_response_text": GOOD_RESPONSE,
        },
    ),
    # schema://name@token form -- channel=webhook is implied
    (
        "pushplus://myhook@{}".format(GOOD_TOKEN),
        {
            "instance": NotifyPushplus,
            "privacy_url": "pushplus://myhook@",
            "requests_response_text": GOOD_RESPONSE,
        },
    ),
    # ----------------------------------------------------------------
    # Native PushPlus API URL (parse_native_url support)
    # ----------------------------------------------------------------
    (
        "https://www.pushplus.plus/send?token={}".format(GOOD_TOKEN),
        {
            "instance": NotifyPushplus,
            "requests_response_text": GOOD_RESPONSE,
        },
    ),
    # ----------------------------------------------------------------
    # API application-level failure (HTTP 200 but JSON code != 200)
    # ----------------------------------------------------------------
    (
        "pushplus://{}".format(GOOD_TOKEN),
        {
            "instance": NotifyPushplus,
            "response": False,
            "requests_response_text": BAD_RESPONSE,
        },
    ),
    # Unparsable JSON body is treated as a failure
    (
        "pushplus://{}".format(GOOD_TOKEN),
        {
            "instance": NotifyPushplus,
            "response": False,
            "requests_response_text": "{bad json",
        },
    ),
    # None body is treated as a failure
    (
        "pushplus://{}".format(GOOD_TOKEN),
        {
            "instance": NotifyPushplus,
            "response": False,
            "requests_response_text": None,
        },
    ),
    # ----------------------------------------------------------------
    # HTTP-level failures
    # ----------------------------------------------------------------
    (
        "pushplus://{}".format(GOOD_TOKEN),
        {
            "instance": NotifyPushplus,
            "response": False,
            "requests_response_code": requests.codes.internal_server_error,
        },
    ),
    (
        "pushplus://{}".format(GOOD_TOKEN),
        {
            "instance": NotifyPushplus,
            "response": False,
            "requests_response_code": 999,
        },
    ),
    # ----------------------------------------------------------------
    # Connection / socket exceptions
    # ----------------------------------------------------------------
    (
        "pushplus://{}".format(GOOD_TOKEN),
        {
            "instance": NotifyPushplus,
            "test_requests_exceptions": True,
        },
    ),
)


def test_plugin_pushplus_urls():
    """NotifyPushplus() Apprise URL test suite."""
    AppriseURLTester(tests=apprise_url_tests).run_all()


def test_plugin_pushplus_init():
    """NotifyPushplus() initialisation and parameter handling."""

    # Minimal valid instantiation -- no targets
    obj = NotifyPushplus(token=GOOD_TOKEN)
    assert obj.token == GOOD_TOKEN
    assert obj.topics == []
    assert obj.channel == PUSHPLUS_CHANNEL_DEFAULT
    assert obj.webhook is None
    assert obj.invalid_targets == []

    # Single group topic via targets list (no prefix required)
    obj = NotifyPushplus(token=GOOD_TOKEN, targets=["mygroup"])
    assert obj.topics == ["mygroup"]
    assert obj.channel == PUSHPLUS_CHANNEL_DEFAULT

    # Multiple group topics
    obj = NotifyPushplus(
        token=GOOD_TOKEN, targets=["group1", "group2", "group3"]
    )
    assert obj.topics == ["group1", "group2", "group3"]

    # Explicit channel via channel= argument
    obj = NotifyPushplus(token=GOOD_TOKEN, channel="mail")
    assert obj.channel == PushPlusChannel.MAIL

    # wecom friendly alias normalises to the "cp" API value
    obj = NotifyPushplus(token=GOOD_TOKEN, channel="wecom")
    assert obj.channel == PushPlusChannel.WECOM
    assert obj.channel == "cp"

    # cp is the canonical API value (same result as wecom)
    obj = NotifyPushplus(token=GOOD_TOKEN, channel="cp")
    assert obj.channel == "cp"

    # Channel is case-insensitive
    obj = NotifyPushplus(token=GOOD_TOKEN, channel="MAIL")
    assert obj.channel == PushPlusChannel.MAIL

    # Invalid channel raises TypeError
    import pytest

    with pytest.raises(TypeError):
        NotifyPushplus(token=GOOD_TOKEN, channel="invalid_channel")

    # Invalid target (bad characters) goes to invalid_targets
    obj = NotifyPushplus(token=GOOD_TOKEN, targets=["bad!target"])
    assert obj.topics == []
    assert obj.invalid_targets == ["bad!target"]

    # Invalid token raises TypeError
    with pytest.raises(TypeError):
        NotifyPushplus(token="short")

    with pytest.raises(TypeError):
        NotifyPushplus(token="bad!token" + "0" * 30)

    # Webhook value stored as None when falsy
    obj = NotifyPushplus(token=GOOD_TOKEN, webhook="")
    assert obj.webhook is None

    # Webhook value stored when truthy
    obj = NotifyPushplus(
        token=GOOD_TOKEN,
        channel="webhook",
        webhook="myhook",
    )
    assert obj.webhook == "myhook"
    assert obj.channel == PushPlusChannel.WEBHOOK


def test_plugin_pushplus_schema_aliases():
    """NotifyPushplus() wecom:// schema alias."""

    # wecom:// schema sets channel=cp (WeCom API value)
    result = NotifyPushplus.parse_url("wecom://{}".format(GOOD_TOKEN))
    assert result is not None
    obj = NotifyPushplus(**result)
    assert obj.channel == PushPlusChannel.WECOM
    assert obj.channel == "cp"

    # An explicit channel= overrides the schema-implied channel
    result = NotifyPushplus.parse_url(
        "wecom://{}?channel=mail".format(GOOD_TOKEN)
    )
    assert result is not None
    obj = NotifyPushplus(**result)
    assert obj.channel == PushPlusChannel.MAIL


def test_plugin_pushplus_all_channels():
    """NotifyPushplus() accepts every documented API channel value."""

    for ch in PUSHPLUS_CHANNELS:
        # Each channel must initialise without raising
        obj = NotifyPushplus(token=GOOD_TOKEN, channel=ch)
        assert obj.channel == ch

        # Channel matching must be case-insensitive
        obj2 = NotifyPushplus(token=GOOD_TOKEN, channel=ch.upper())
        assert obj2.channel == ch


def test_plugin_pushplus_url():
    """NotifyPushplus() url() output and round-trip fidelity."""

    # Personal notification URL -- no topics in path, default channel omitted
    obj = NotifyPushplus(token=GOOD_TOKEN)
    url = obj.url()
    assert url.startswith("pushplus://")
    # Token appears in plain text
    assert GOOD_TOKEN in url
    # No topic path segment
    assert "/mygroup" not in url
    # Default channel is omitted from params
    assert "channel=" not in url

    # Privacy URL must mask the token
    priv = obj.url(privacy=True)
    assert GOOD_TOKEN not in priv
    assert "****" in priv

    # Group topic URL -- topic appears in the path
    obj = NotifyPushplus(token=GOOD_TOKEN, targets=["mygroup"])
    url = obj.url()
    assert "/mygroup/" in url
    priv = obj.url(privacy=True)
    assert "/mygroup/" in priv
    assert GOOD_TOKEN not in priv

    # Multiple topics all appear in the path
    obj = NotifyPushplus(token=GOOD_TOKEN, targets=["grp1", "grp2"])
    url = obj.url()
    assert "grp1" in url
    assert "grp2" in url

    # Non-default channel appears as ?channel= query parameter
    obj = NotifyPushplus(token=GOOD_TOKEN, channel="mail")
    url = obj.url()
    assert "channel=mail" in url

    # Default channel (wechat) is suppressed from the URL
    obj = NotifyPushplus(token=GOOD_TOKEN, channel="wechat")
    url = obj.url()
    assert "channel=" not in url

    # When channel=webhook with a name, url() uses schema://name@token form
    obj = NotifyPushplus(
        token=GOOD_TOKEN,
        channel="webhook",
        webhook="myhook",
    )
    url = obj.url()
    assert "pushplus://myhook@" in url
    # channel= and name= are both implied by user@ -- omitted from params
    assert "channel=" not in url
    assert "name=" not in url

    # Privacy URL preserves the webhook name but masks the token
    priv = obj.url(privacy=True)
    assert "pushplus://myhook@" in priv
    assert GOOD_TOKEN not in priv
    assert "****" in priv

    # Webhook prefix form with group topics -- schema://name@token/topic/ form
    obj = NotifyPushplus(
        token=GOOD_TOKEN,
        channel="webhook",
        webhook="myhook",
        targets=["mygroup"],
    )
    url = obj.url()
    assert "pushplus://myhook@" in url
    assert "mygroup" in url
    assert "channel=" not in url
    assert "name=" not in url

    # Webhook name suppressed when channel is not webhook
    obj = NotifyPushplus(
        token=GOOD_TOKEN,
        channel="mail",
        webhook="myhook",
    )
    url = obj.url()
    assert "name=" not in url
    assert "@" not in url

    # Webhook channel without a name uses the plain ?channel=webhook form
    obj = NotifyPushplus(token=GOOD_TOKEN, channel="webhook")
    url = obj.url()
    assert "name=" not in url
    assert "channel=webhook" in url
    assert "@" not in url

    # Invalid targets are preserved in the URL path for round-trip fidelity
    obj = NotifyPushplus(token=GOOD_TOKEN, targets=["bad!target"])
    url = obj.url()
    assert "bad" in url

    # url_identifier contains only the schema and token (no targets or channel)
    obj = NotifyPushplus(token=GOOD_TOKEN, targets=["grp"])
    assert obj.url_identifier == ("pushplus", GOOD_TOKEN)

    # Schema alias (wecom://) always normalises back to pushplus://
    result = NotifyPushplus.parse_url("wecom://{}".format(GOOD_TOKEN))
    obj = NotifyPushplus(**result)
    url = obj.url()
    assert url.startswith("pushplus://")
    assert "channel=cp" in url


@mock.patch("requests.post")
def test_plugin_pushplus_send_ok(mock_post):
    """NotifyPushplus() successful personal-notification send."""

    # Prepare a successful mock response
    response = mock.Mock()
    response.status_code = requests.codes.ok
    response.content = GOOD_RESPONSE.encode("utf-8")
    mock_post.return_value = response

    # Personal notification (no topic) -- single API call
    obj = NotifyPushplus(token=GOOD_TOKEN)
    assert obj.send(body="hello", title="title") is True
    assert mock_post.call_count == 1

    # Verify the endpoint URL
    call = mock_post.call_args
    assert call[0][0] == "https://www.pushplus.plus/send"

    # Verify the core payload fields
    payload = loads(call[1]["data"])
    assert payload["token"] == GOOD_TOKEN
    assert payload["content"] == "hello"
    assert payload["title"] == "title"
    # Default format is HTML -- PushPlus template should be "html"
    assert payload["template"] == PUSHPLUS_FORMAT_MAP[NotifyFormat.HTML]
    assert payload["channel"] == PUSHPLUS_CHANNEL_DEFAULT
    # topic and webhook keys must be absent when not configured
    assert "topic" not in payload
    assert "webhook" not in payload

    mock_post.reset_mock()
    response.content = GOOD_RESPONSE.encode("utf-8")

    # When no title is supplied the body is used as the title fallback
    assert obj.send(body="only body") is True
    payload = loads(mock_post.call_args[1]["data"])
    assert payload["title"] == "only body"
    assert payload["content"] == "only body"


@mock.patch("requests.post")
def test_plugin_pushplus_send_formats(mock_post):
    """NotifyPushplus() maps notify_format to the PushPlus template field."""

    response = mock.Mock()
    response.status_code = requests.codes.ok
    response.content = GOOD_RESPONSE.encode("utf-8")
    mock_post.return_value = response

    obj = NotifyPushplus(token=GOOD_TOKEN)

    # Markdown format -> "markdown" template
    obj.notify_format = NotifyFormat.MARKDOWN
    assert obj.send(body="# heading") is True
    payload = loads(mock_post.call_args[1]["data"])
    assert payload["template"] == "markdown"

    mock_post.reset_mock()
    response.content = GOOD_RESPONSE.encode("utf-8")

    # Text format -> "txt" template
    obj.notify_format = NotifyFormat.TEXT
    assert obj.send(body="plain text") is True
    payload = loads(mock_post.call_args[1]["data"])
    assert payload["template"] == "txt"

    mock_post.reset_mock()
    response.content = GOOD_RESPONSE.encode("utf-8")

    # HTML format -> "html" template
    obj.notify_format = NotifyFormat.HTML
    assert obj.send(body="<b>bold</b>") is True
    payload = loads(mock_post.call_args[1]["data"])
    assert payload["template"] == "html"


@mock.patch("requests.post")
def test_plugin_pushplus_send_topic(mock_post):
    """NotifyPushplus() includes the topic in the payload when configured."""

    response = mock.Mock()
    response.status_code = requests.codes.ok
    response.content = GOOD_RESPONSE.encode("utf-8")
    mock_post.return_value = response

    # Single topic -- one API call, topic present in payload
    obj = NotifyPushplus(token=GOOD_TOKEN, targets=["mygroup"])
    assert obj.send(body="group msg") is True
    assert mock_post.call_count == 1
    payload = loads(mock_post.call_args[1]["data"])
    assert payload["topic"] == "mygroup"


@mock.patch("requests.post")
def test_plugin_pushplus_send_multiple_topics(mock_post):
    """NotifyPushplus() makes one API call per group topic."""

    response = mock.Mock()
    response.status_code = requests.codes.ok
    response.content = GOOD_RESPONSE.encode("utf-8")
    mock_post.return_value = response

    # Three topics -> three separate API calls
    obj = NotifyPushplus(
        token=GOOD_TOKEN, targets=["topic1", "topic2", "topic3"]
    )
    assert len(obj.topics) == 3
    assert obj.send(body="multi msg") is True
    assert mock_post.call_count == 3

    # Verify each call carries the correct topic value
    calls = mock_post.call_args_list
    topics_sent = [loads(c[1]["data"])["topic"] for c in calls]
    assert topics_sent == ["topic1", "topic2", "topic3"]

    mock_post.reset_mock()
    response.content = GOOD_RESPONSE.encode("utf-8")

    # Topics with a channel override -- all API calls use the specified channel
    obj2 = NotifyPushplus(
        token=GOOD_TOKEN, targets=["grp1", "grp2"], channel="mail"
    )
    assert len(obj2.topics) == 2
    assert obj2.channel == PushPlusChannel.MAIL
    assert obj2.send(body="mixed") is True
    assert mock_post.call_count == 2
    for call in mock_post.call_args_list:
        assert loads(call[1]["data"])["channel"] == "mail"

    mock_post.reset_mock()

    # Partial failure: first topic succeeds, second fails -> overall False
    good_resp = mock.Mock()
    good_resp.status_code = requests.codes.ok
    good_resp.content = GOOD_RESPONSE.encode("utf-8")
    bad_resp = mock.Mock()
    bad_resp.status_code = requests.codes.ok
    bad_resp.content = BAD_RESPONSE.encode("utf-8")
    mock_post.side_effect = [good_resp, bad_resp]

    obj3 = NotifyPushplus(token=GOOD_TOKEN, targets=["grp1", "grp2"])
    assert obj3.send(body="partial") is False
    assert mock_post.call_count == 2


@mock.patch("requests.post")
def test_plugin_pushplus_send_webhook(mock_post):
    """NotifyPushplus() webhook channel behaviour."""

    response = mock.Mock()
    response.status_code = requests.codes.ok
    response.content = GOOD_RESPONSE.encode("utf-8")
    mock_post.return_value = response

    # Webhook channel with a named endpoint -- webhook key in payload
    obj = NotifyPushplus(
        token=GOOD_TOKEN,
        channel="webhook",
        webhook="myhook",
    )
    assert obj.send(body="hook msg") is True
    payload = loads(mock_post.call_args[1]["data"])
    assert payload["channel"] == PushPlusChannel.WEBHOOK
    assert payload["webhook"] == "myhook"

    mock_post.reset_mock()
    response.content = GOOD_RESPONSE.encode("utf-8")

    # Webhook channel without a name -- webhook key must be absent
    obj2 = NotifyPushplus(token=GOOD_TOKEN, channel="webhook")
    assert obj2.send(body="hook no name") is True
    payload2 = loads(mock_post.call_args[1]["data"])
    assert "webhook" not in payload2

    mock_post.reset_mock()
    response.content = GOOD_RESPONSE.encode("utf-8")

    # Non-webhook channel with a webhook value set -- key must be absent
    obj3 = NotifyPushplus(
        token=GOOD_TOKEN,
        channel="mail",
        webhook="myhook",
    )
    assert obj3.send(body="mail msg") is True
    payload3 = loads(mock_post.call_args[1]["data"])
    assert "webhook" not in payload3


@mock.patch("requests.post")
def test_plugin_pushplus_send_http_error(mock_post):
    """NotifyPushplus() handles HTTP-level errors correctly."""

    response = mock.Mock()
    response.content = b""
    mock_post.return_value = response

    obj = NotifyPushplus(token=GOOD_TOKEN)

    # HTTP 500
    response.status_code = requests.codes.internal_server_error
    assert obj.send(body="msg") is False

    # Unknown HTTP status code
    response.status_code = 999
    assert obj.send(body="msg") is False


@mock.patch("requests.post")
def test_plugin_pushplus_send_api_error(mock_post):
    """NotifyPushplus() handles API-level failures (HTTP 200, bad code)."""

    response = mock.Mock()
    response.status_code = requests.codes.ok
    mock_post.return_value = response

    obj = NotifyPushplus(token=GOOD_TOKEN)

    # Known API error code (907 = Token does not exist)
    response.content = BAD_RESPONSE.encode("utf-8")
    assert obj.send(body="msg") is False

    # Unknown API error code with a msg field -- falls back to msg
    response.content = dumps({"code": 999, "msg": "some error"}).encode(
        "utf-8"
    )
    assert obj.send(body="msg") is False

    # Unknown API error code with no msg field -- generic fallback
    response.content = dumps({"code": 999}).encode("utf-8")
    assert obj.send(body="msg") is False

    # Unparsable JSON body
    response.content = b"{bad json"
    assert obj.send(body="msg") is False

    # None content -- treated as unparsable
    response.content = None
    assert obj.send(body="msg") is False


@mock.patch("requests.post")
def test_plugin_pushplus_send_exception(mock_post):
    """NotifyPushplus() handles connection exceptions gracefully."""
    mock_post.side_effect = requests.RequestException("connection error")
    obj = NotifyPushplus(token=GOOD_TOKEN)
    assert obj.send(body="msg") is False


@mock.patch("requests.post")
def test_plugin_pushplus_send_exception_multi_topic(mock_post):
    """NotifyPushplus() continues after a connection error on one topic."""

    # First call raises, second succeeds -- overall result is False
    good_resp = mock.Mock()
    good_resp.status_code = requests.codes.ok
    good_resp.content = GOOD_RESPONSE.encode("utf-8")
    mock_post.side_effect = [
        requests.RequestException("network error"),
        good_resp,
    ]

    obj = NotifyPushplus(token=GOOD_TOKEN, targets=["grp1", "grp2"])
    assert obj.send(body="msg") is False
    # Both topics must still be attempted
    assert mock_post.call_count == 2


def test_plugin_pushplus_parse_url():
    """NotifyPushplus() parse_url() extracts all supported fields."""

    # Token in the host position
    result = NotifyPushplus.parse_url("pushplus://{}".format(GOOD_TOKEN))
    assert result["token"] == GOOD_TOKEN
    assert result["targets"] == []

    # Token supplied as a ?token= query parameter
    result = NotifyPushplus.parse_url(
        "pushplus://?token={}".format(GOOD_TOKEN)
    )
    assert result["token"] == GOOD_TOKEN

    # Topic in the URL path becomes a plain targets entry
    result = NotifyPushplus.parse_url(
        "pushplus://{}/mygroup".format(GOOD_TOKEN)
    )
    assert "mygroup" in result["targets"]

    # Two topics in the path
    result = NotifyPushplus.parse_url(
        "pushplus://{}/grp1/grp2".format(GOOD_TOKEN)
    )
    assert "grp1" in result["targets"]
    assert "grp2" in result["targets"]

    # Backward compat: ?topic= adds a plain topic entry
    result = NotifyPushplus.parse_url(
        "pushplus://{}/?topic=mygroup".format(GOOD_TOKEN)
    )
    assert "mygroup" in result["targets"]

    # ?to= alias appends additional targets
    result = NotifyPushplus.parse_url(
        "pushplus://{}/?to=mygroup".format(GOOD_TOKEN)
    )
    assert "mygroup" in result["targets"]

    # Delivery channel via ?channel=
    result = NotifyPushplus.parse_url(
        "pushplus://{}?channel=mail".format(GOOD_TOKEN)
    )
    assert result.get("channel") == "mail"

    # Delivery channel via ?mode= alias
    result = NotifyPushplus.parse_url(
        "pushplus://{}?mode=sms".format(GOOD_TOKEN)
    )
    assert result.get("channel") == "sms"

    # channel= takes precedence over mode= when both are supplied
    result = NotifyPushplus.parse_url(
        "pushplus://{}?mode=sms&channel=mail".format(GOOD_TOKEN)
    )
    assert result.get("channel") == "mail"

    # Webhook name extracted into the "webhook" key from ?name=
    result = NotifyPushplus.parse_url(
        "pushplus://{}?channel=webhook&name=myhook".format(GOOD_TOKEN)
    )
    assert result.get("webhook") == "myhook"
    assert result.get("channel") == "webhook"

    # schema://name@token form: user@ -> webhook, implies channel=webhook
    result = NotifyPushplus.parse_url(
        "pushplus://myhook@{}".format(GOOD_TOKEN)
    )
    assert result.get("webhook") == "myhook"
    assert result.get("channel") == PushPlusChannel.WEBHOOK

    # Explicit ?channel= overrides the webhook implication from user@
    result = NotifyPushplus.parse_url(
        "pushplus://myhook@{}?channel=mail".format(GOOD_TOKEN)
    )
    assert result.get("webhook") == "myhook"
    assert result.get("channel") == "mail"

    # ?name= takes precedence over user@ when both are supplied
    result = NotifyPushplus.parse_url(
        "pushplus://other@{}?channel=webhook&name=myhook".format(GOOD_TOKEN)
    )
    assert result.get("webhook") == "myhook"
    assert result.get("channel") == "webhook"


def test_plugin_pushplus_parse_native_url():
    """NotifyPushplus() parse_native_url() handles all edge cases."""

    # Basic native URL -- token extracted correctly
    result = NotifyPushplus.parse_native_url(
        "https://www.pushplus.plus/send?token={}".format(GOOD_TOKEN)
    )
    assert result is not None
    obj = NotifyPushplus(**result)
    assert obj.token == GOOD_TOKEN

    # Native URL with topic parameter preserved via round-trip
    result = NotifyPushplus.parse_native_url(
        "https://www.pushplus.plus/send?token={}&topic=mygroup".format(
            GOOD_TOKEN
        )
    )
    assert result is not None
    obj = NotifyPushplus(**result)
    assert "mygroup" in obj.topics

    # Native URL with channel parameter preserved
    result = NotifyPushplus.parse_native_url(
        "https://www.pushplus.plus/send?token={}&channel=mail".format(
            GOOD_TOKEN
        )
    )
    assert result is not None
    obj = NotifyPushplus(**result)
    assert obj.channel == PushPlusChannel.MAIL

    # Native URL with webhook name parameter preserved
    result = NotifyPushplus.parse_native_url(
        "https://www.pushplus.plus/send"
        "?token={}&channel=webhook&name=myhook".format(GOOD_TOKEN)
    )
    assert result is not None
    obj = NotifyPushplus(**result)
    assert obj.webhook == "myhook"
    assert obj.channel == PushPlusChannel.WEBHOOK

    # Native URL without a token -- not matched, returns None
    assert (
        NotifyPushplus.parse_native_url(
            "https://www.pushplus.plus/send?channel=mail"
        )
        is None
    )

    # Completely different domain -- not matched, returns None
    assert (
        NotifyPushplus.parse_native_url(
            "https://other.example.com/send?token={}".format(GOOD_TOKEN)
        )
        is None
    )

    # Native URL with no query string -- not matched, returns None
    assert (
        NotifyPushplus.parse_native_url("https://www.pushplus.plus/send")
        is None
    )


@mock.patch("requests.post")
def test_plugin_pushplus_apprise_integration(mock_post):
    """NotifyPushplus() integrates correctly with the Apprise object."""

    response = mock.Mock()
    response.status_code = requests.codes.ok
    response.content = GOOD_RESPONSE.encode("utf-8")
    mock_post.return_value = response

    # Personal URL -- single notification
    aobj = Apprise()
    assert aobj.add("pushplus://{}".format(GOOD_TOKEN))
    assert len(aobj) == 1
    assert aobj.notify(title="Test Title", body="Test Body") is True
    assert mock_post.call_count == 1

    mock_post.reset_mock()
    response.content = GOOD_RESPONSE.encode("utf-8")

    # Group URL with markdown format override
    aobj2 = Apprise()
    assert aobj2.add(
        "pushplus://{}/mygroup?format=markdown".format(GOOD_TOKEN)
    )
    assert aobj2.notify(body="Group message") is True
    assert mock_post.call_count == 1
    payload = loads(mock_post.call_args[1]["data"])
    # format=markdown must map to the PushPlus "markdown" template
    assert payload["template"] == "markdown"
    assert payload["topic"] == "mygroup"

    mock_post.reset_mock()
    response.content = GOOD_RESPONSE.encode("utf-8")

    # wecom:// schema alias -- channel=cp in payload
    aobj3 = Apprise()
    assert aobj3.add("wecom://{}".format(GOOD_TOKEN))
    assert aobj3.notify(body="WeCom message") is True
    payload = loads(mock_post.call_args[1]["data"])
    assert payload["channel"] == PushPlusChannel.WECOM



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Disable logging for a cleaner testing output
import logging
import os
from unittest import mock

from helpers import AppriseURLTester
import pytest
import requests

from apprise import Apprise, AppriseAttachment, NotifyType
from apprise.plugins.postmark import NotifyPostmark

logging.disable(logging.CRITICAL)

# Attachment Directory
TEST_VAR_DIR = os.path.join(os.path.dirname(__file__), "var")

# Our Testing URLs
apprise_url_tests = (
    (
        "postmark://",
        {
            # No credentials
            "instance": TypeError,
        },
    ),
    (
        "postmark://:@/",
        {
            # Empty credentials
            "instance": TypeError,
        },
    ),
    (
        "postmark://abcd",
        {
            # API key only, no from email
            "instance": TypeError,
        },
    ),
    (
        "postmark://abcd@host",
        {
            # API key only, no from email
            "instance": TypeError,
        },
    ),
    (
        "postmark://abcd:user@example.com",
        {
            # No targets; defaults to from address
            "instance": NotifyPostmark,
        },
    ),
    (
        "postmark://abcd:user@example.com?format=text",
        {
            # Text format override
            "instance": NotifyPostmark,
        },
    ),
    (
        "postmark://abcd:user@example.com/newuser1@example.com",
        {
            # A good email with explicit target
            "instance": NotifyPostmark,
            "force_debug": True,
        },
    ),
    (
        "postmark://abcd:user@example.com/newuser2@example.com?name=John",
        {
            # With From display name
            "instance": NotifyPostmark,
            "privacy_url": (
                "postmark://a...d:user@example.com/newuser2@example.com"
            ),
            "url_matches": r"name=John",
        },
    ),
    (
        (
            "postmark://abcd:user@example.com/newuser3@example.com"
            "?reply=reply@example.com"
        ),
        {
            # With Reply-To
            "instance": NotifyPostmark,
            "url_matches": r"reply=",
        },
    ),
    (
        (
            "postmark://abcd:user@example.com/newuser4@example.com"
            "?bcc=bcc@nuxref.com"
        ),
        {
            # With Blind Carbon Copy
            "instance": NotifyPostmark,
        },
    ),
    (
        (
            "postmark://abcd:user@example.com/newuser5@example.com"
            "?cc=cc@nuxref.com"
        ),
        {
            # With Carbon Copy
            "instance": NotifyPostmark,
        },
    ),
    (
        (
            "postmark://abcd:user@example.com/newuser5@example.com"
            "?cc=Chris<cc@nuxref.com>"
        ),
        {
            # With named Carbon Copy
            "instance": NotifyPostmark,
        },
    ),
    (
        (
            "postmark://abcd:user@example.com/newuser6@example.com"
            "?to=extra@nuxref.com"
        ),
        {
            # Via ?to= for additional targets
            "instance": NotifyPostmark,
        },
    ),
    (
        "postmark://abcd:user@example.com/bademailaddress",
        {
            # Invalid target dropped; no valid recipients -- fails to send
            "instance": NotifyPostmark,
            "notify_response": False,
        },
    ),
    (
        "postmark://abcd:user@example.ca/newuser@example.ca",
        {
            "instance": NotifyPostmark,
            # force a failure
            "response": False,
            "requests_response_code": requests.codes.internal_server_error,
        },
    ),
    (
        "postmark://abcd:user@example.uk/newuser@example.uk",
        {
            "instance": NotifyPostmark,
            # throw a bizarre code forcing us to fail to look it up
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        "postmark://abcd:user@example.au/newuser@example.au",
        {
            "instance": NotifyPostmark,
            # Throws a series of connection exceptions with this flag
            "test_requests_exceptions": True,
        },
    ),
)


def test_plugin_postmark_urls():
    """NotifyPostmark() Apprise URLs."""

    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()


@mock.patch("requests.get")
@mock.patch("requests.post")
def test_plugin_postmark_edge_cases(mock_post, mock_get):
    """NotifyPostmark() Edge Cases."""

    # No API key
    with pytest.raises(TypeError):
        NotifyPostmark(apikey=None, from_email="user@example.com")

    # Invalid from email
    with pytest.raises(TypeError):
        NotifyPostmark(apikey="abcd", from_email="!invalid")

    # No from email (None)
    with pytest.raises(TypeError):
        NotifyPostmark(apikey="abcd", from_email=None)

    # Invalid target email -- plugin loads but target is dropped
    obj = NotifyPostmark(
        apikey="abcd",
        from_email="user@example.com",
        targets="!invalid",
    )
    # All targets dropped -> send() returns False
    assert isinstance(obj, NotifyPostmark)

    # Test invalid bcc/cc entries mixed with valid ones
    assert isinstance(
        NotifyPostmark(
            apikey="abcd",
            from_email="l2g@example.com",
            bcc=("abc@def.com", "!invalid"),
            cc=("abc@test.org", "!invalid"),
        ),
        NotifyPostmark,
    )

    # Test invalid reply_to entries mixed with valid ones
    assert isinstance(
        NotifyPostmark(
            apikey="abcd",
            from_email="l2g@example.com",
            reply_to=("reply@def.com", "!invalid"),
        ),
        NotifyPostmark,
    )


@mock.patch("requests.get")
@mock.patch("requests.post")
def test_plugin_postmark_send(mock_post, mock_get):
    """NotifyPostmark() Send."""

    def _mk_resp(code=requests.codes.ok):
        r = mock.Mock()
        r.status_code = code
        r.content = b""
        return r

    mock_post.return_value = _mk_resp()
    mock_get.return_value = _mk_resp()

    # Basic send to self (no explicit target)
    obj = Apprise.instantiate("postmark://abcd:user@example.com")
    assert isinstance(obj, NotifyPostmark)
    assert obj.notify(body="body", title="title") is True

    # Verify the X-Postmark-Server-Token header was set
    assert mock_post.call_count == 1
    call_kwargs = mock_post.call_args
    import json

    posted = json.loads(call_kwargs[1]["data"])
    assert posted["To"] == "user@example.com"
    assert "X-Postmark-Server-Token" in call_kwargs[1]["headers"]

    mock_post.reset_mock()

    # Send with an explicit target
    obj = Apprise.instantiate(
        "postmark://abcd:user@example.com/target@example.com"
    )
    assert obj.notify(body="body", title="title") is True
    posted = json.loads(mock_post.call_args[1]["data"])
    assert posted["To"] == "target@example.com"
    mock_post.reset_mock()

    # Send with cc and bcc
    obj = NotifyPostmark(
        apikey="abcd",
        from_email="user@example.com",
        targets=["target@example.com"],
        cc=["cc@example.com"],
        bcc=["bcc@example.com"],
    )
    assert obj.notify(body="body", title="title") is True
    posted = json.loads(mock_post.call_args[1]["data"])
    assert "Cc" in posted
    assert "Bcc" in posted
    mock_post.reset_mock()

    # Send with reply_to
    obj = NotifyPostmark(
        apikey="abcd",
        from_email="user@example.com",
        reply_to=["reply@example.com"],
    )
    assert obj.notify(body="body", title="title") is True
    posted = json.loads(mock_post.call_args[1]["data"])
    assert "ReplyTo" in posted
    mock_post.reset_mock()

    # Empty title falls back to default_empty_subject
    obj = NotifyPostmark(
        apikey="abcd",
        from_email="user@example.com",
    )
    assert obj.notify(body="body", title="") is True
    posted = json.loads(mock_post.call_args[1]["data"])
    assert posted["Subject"] == NotifyPostmark.default_empty_subject
    mock_post.reset_mock()

    # Text format
    obj = NotifyPostmark(
        apikey="abcd",
        from_email="user@example.com",
    )
    from apprise.common import NotifyFormat

    obj.notify_format = NotifyFormat.TEXT
    assert obj.notify(body="body", title="title") is True
    posted = json.loads(mock_post.call_args[1]["data"])
    assert "TextBody" in posted
    assert "HtmlBody" not in posted
    mock_post.reset_mock()

    # HTTP error response
    mock_post.return_value = _mk_resp(requests.codes.internal_server_error)
    obj = Apprise.instantiate(
        "postmark://abcd:user@example.com/target@example.com"
    )
    assert obj.notify(body="body", title="title") is False
    mock_post.reset_mock()

    # Unknown HTTP code
    mock_post.return_value = _mk_resp(999)
    assert obj.notify(body="body", title="title") is False
    mock_post.reset_mock()

    # RequestException
    mock_post.side_effect = requests.RequestException("Connection failed")
    assert obj.notify(body="body", title="title") is False
    mock_post.side_effect = None
    mock_post.reset_mock()

    # No valid targets -> False without any HTTP call
    obj = NotifyPostmark(
        apikey="abcd",
        from_email="user@example.com",
        targets=["!invalid"],
    )
    mock_post.return_value = _mk_resp()
    assert obj.notify(body="body", title="title") is False
    assert mock_post.call_count == 0

    # Multiple targets -- partial failure
    mock_post.reset_mock()
    responses = [_mk_resp(), _mk_resp(requests.codes.bad_request)]
    mock_post.side_effect = responses
    obj = NotifyPostmark(
        apikey="abcd",
        from_email="user@example.com",
        targets=["t1@example.com", "t2@example.com"],
    )
    assert obj.notify(body="body", title="title") is False
    assert mock_post.call_count == 2
    mock_post.side_effect = None


@mock.patch("requests.post")
def test_plugin_postmark_attachments(mock_post):
    """NotifyPostmark() Attachments."""

    def _mk_resp(code=requests.codes.ok):
        r = mock.Mock()
        r.status_code = code
        r.content = b""
        return r

    mock_post.return_value = _mk_resp()

    # Single attachment success
    path = os.path.join(TEST_VAR_DIR, "apprise-test.gif")
    attach = AppriseAttachment(path)
    obj = Apprise.instantiate("postmark://abcd:user@example.com")
    assert isinstance(obj, NotifyPostmark)
    assert (
        obj.notify(
            body="body",
            title="title",
            notify_type=NotifyType.INFO,
            attach=attach,
        )
        is True
    )

    import json

    posted = json.loads(mock_post.call_args[1]["data"])
    assert "Attachments" in posted
    assert len(posted["Attachments"]) == 1
    assert posted["Attachments"][0]["Name"] == "apprise-test.gif"
    mock_post.reset_mock()

    # Multiple attachments
    attach = AppriseAttachment(os.path.join(TEST_VAR_DIR, "apprise-test.gif"))
    attach.add(os.path.join(TEST_VAR_DIR, "apprise-test.png"))
    assert (
        obj.notify(
            body="body",
            title="title",
            notify_type=NotifyType.INFO,
            attach=attach,
        )
        is True
    )
    posted = json.loads(mock_post.call_args[1]["data"])
    assert len(posted["Attachments"]) == 2
    mock_post.reset_mock()

    # Attachment inaccessible (file not found)
    with mock.patch("os.path.isfile", return_value=False):
        assert (
            obj.notify(
                body="body",
                title="title",
                notify_type=NotifyType.INFO,
                attach=AppriseAttachment(
                    os.path.join(TEST_VAR_DIR, "apprise-test.gif")
                ),
            )
            is False
        )

    # Attachment read error (OSError)
    with mock.patch("builtins.open", side_effect=OSError):
        assert (
            obj.notify(
                body="body",
                title="title",
                notify_type=NotifyType.INFO,
                attach=AppriseAttachment(
                    os.path.join(TEST_VAR_DIR, "apprise-test.gif")
                ),
            )
            is False
        )


@mock.patch("requests.post")
def test_plugin_postmark_url_parsing(mock_post):
    """NotifyPostmark() URL parsing and round-trip."""

    def _mk_resp(code=requests.codes.ok):
        r = mock.Mock()
        r.status_code = code
        r.content = b""
        return r

    mock_post.return_value = _mk_resp()

    # Basic round-trip
    obj = NotifyPostmark(
        apikey="mytoken",
        from_email="user@example.com",
    )
    url = obj.url()
    result = NotifyPostmark.parse_url(url)
    assert result is not None
    obj2 = NotifyPostmark(**result)
    assert obj.url_identifier == obj2.url_identifier

    # With explicit target
    obj = NotifyPostmark(
        apikey="mytoken",
        from_email="user@example.com",
        targets=["target@example.com"],
    )
    url = obj.url()
    result = NotifyPostmark.parse_url(url)
    assert result is not None
    obj2 = NotifyPostmark(**result)
    assert len(obj.targets) == len(obj2.targets)

    # With from_name
    obj = NotifyPostmark(
        apikey="mytoken",
        from_email="user@example.com",
        from_name="Alice",
    )
    url = obj.url()
    assert "name=Alice" in url
    result = NotifyPostmark.parse_url(url)
    assert result is not None
    obj2 = NotifyPostmark(**result)
    assert obj2.from_addr[0] == "Alice"

    # With bcc and cc
    obj = NotifyPostmark(
        apikey="mytoken",
        from_email="user@example.com",
        targets=["target@example.com"],
        cc=["cc@example.com"],
        bcc=["bcc@example.com"],
    )
    url = obj.url()
    result = NotifyPostmark.parse_url(url)
    assert result is not None
    obj2 = NotifyPostmark(**result)
    assert "cc@example.com" in obj2.cc
    assert "bcc@example.com" in obj2.bcc

    # With reply_to
    obj = NotifyPostmark(
        apikey="mytoken",
        from_email="user@example.com",
        reply_to=["reply@example.com"],
    )
    url = obj.url()
    result = NotifyPostmark.parse_url(url)
    assert result is not None
    obj2 = NotifyPostmark(**result)
    assert "reply@example.com" in obj2.reply_to

    # parse_url returns None for completely unparseable input
    assert NotifyPostmark.parse_url(None) is None

    # ?apikey= override
    result = NotifyPostmark.parse_url(
        "postmark://?apikey=overridekey&from=sender@example.com"
        "&to=target@example.com"
    )
    assert result is not None
    assert result["apikey"] == "overridekey"

    # ?from= override (from address via query param)
    result = NotifyPostmark.parse_url(
        "postmark://tok@example.com/path@example.com?from=sender@example.com"
    )
    assert result is not None
    assert result["from_email"] == "sender@example.com"

    # privacy_url hides the API key
    obj = NotifyPostmark(
        apikey="secrettoken",
        from_email="user@example.com",
    )
    privacy = obj.url(privacy=True)
    assert "secrettoken" not in privacy
    assert "s...n" in privacy or "..." in privacy

    # __len__
    obj = NotifyPostmark(
        apikey="mytoken",
        from_email="user@example.com",
        targets=["a@example.com", "b@example.com"],
    )
    assert len(obj) == 2

    # url_identifier
    obj1 = NotifyPostmark(
        apikey="tok",
        from_email="user@example.com",
    )
    obj2 = NotifyPostmark(
        apikey="tok",
        from_email="user@example.com",
        targets=["other@example.com"],
    )
    assert obj1.url_identifier == obj2.url_identifier



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import gc
import mimetypes
import os
import sys

import pytest

from apprise import (
    AttachmentManager,
    ConfigurationManager,
    NotificationManager,
)

sys.path.append(os.path.join(os.path.dirname(__file__), "helpers"))

# Grant access to our Notification Manager Singleton
N_MGR = NotificationManager()
# Grant access to our Config Manager Singleton
C_MGR = ConfigurationManager()
# Grant access to our Attachment Manager Singleton
A_MGR = AttachmentManager()


@pytest.fixture(scope="function", autouse=True)
def mimetypes_always_available():
    """A pytest session fixture which ensures mimetypes is set correctly
    pointing to our temporary mime.types file."""
    files = (os.path.join(os.path.dirname(__file__), "var", "mime.types"),)
    mimetypes.init(files=files)


@pytest.fixture(scope="function", autouse=True)
def no_throttling_everywhere(session_mocker):
    """A pytest session fixture which disables throttling on all notifiers.

    It is automatically enabled.
    """
    # Ensure we're working with a clean slate for each test
    N_MGR.unload_modules()
    C_MGR.unload_modules()
    A_MGR.unload_modules()

    for plugin in N_MGR.plugins():
        session_mocker.patch.object(plugin, "request_rate_per_sec", 0)


@pytest.fixture(scope="function", autouse=True)
def collect_all_garbage(session_mocker):
    """A pytest session fixture to ensure no __del__ cleanup call from one
    plugin will cause testing issues with another.

    Run garbage collection after every test
    """
    # Force garbage collection
    gc.collect()



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from json import dumps
import logging

from helpers import AppriseURLTester
import requests

from apprise import Apprise
from apprise.plugins.viber import NotifyViber

logging.disable(logging.CRITICAL)

VIBER_GOOD_RESPONSE = dumps({"status": 0, "status_message": "ok"})
VIBER_BAD_RESPONSE = dumps(
    {"status": 12, "status_message": "Too many requests"}
)

# Our Testing URLs
apprise_url_tests = (
    ("viber://", False),
    ("viber:///", False),
    ("viber://tokena", {"instance": NotifyViber, "notify_response": False}),
    (
        "viber://?token=tokenb",
        {"instance": NotifyViber, "notify_response": False},
    ),
    (
        "viber://token/targetx",
        {
            "instance": NotifyViber,
            # Our response expected server response
            "requests_response_text": VIBER_GOOD_RESPONSE,
        },
    ),
    (
        "viber://token/t1/t2?from=Viber%20Bot",
        {
            "instance": NotifyViber,
            "requests_response_text": VIBER_GOOD_RESPONSE,
        },
    ),
    (
        "viber://t1/t2?token=token",
        {
            "instance": NotifyViber,
            "requests_response_text": VIBER_GOOD_RESPONSE,
        },
    ),
    (
        "viber://?token=token&to=t5",
        {
            "instance": NotifyViber,
            "requests_response_text": VIBER_GOOD_RESPONSE,
        },
    ),
    (
        "viber://token/t3?avatar=value",
        {
            "instance": NotifyViber,
            "requests_response_text": VIBER_GOOD_RESPONSE,
        },
    ),
    (
        "viber://token/?to=abc,def",
        {
            "instance": NotifyViber,
            "requests_response_text": VIBER_GOOD_RESPONSE,
        },
    ),
    (
        "viber://token/m12/?from={}".format(
            "a" * (NotifyViber.viber_sender_name_limit + 1)
        ),
        {
            "instance": NotifyViber,
            "requests_response_text": VIBER_GOOD_RESPONSE,
        },
    ),
    (
        "viber://token/m12/?from={}".format(
            "b" * (NotifyViber.viber_sender_name_limit)
        ),
        {
            "instance": NotifyViber,
            "requests_response_text": VIBER_GOOD_RESPONSE,
        },
    ),
    (
        "viber://?token=token&to=hij,klm",
        {
            "instance": NotifyViber,
            "requests_response_text": VIBER_GOOD_RESPONSE,
        },
    ),
    (
        "viber://?token=token&to=nop,qrs",
        {
            "instance": NotifyViber,
            "requests_response_text": VIBER_BAD_RESPONSE,
            "notify_response": False,
        },
    ),
    (
        "viber://?token=token&to=tuv,wxy",
        {
            "instance": NotifyViber,
            # Bad JSON
            "requests_response_text": "{",
            "notify_response": False,
        },
    ),
    # Privacy redaction of token
    (
        "viber://token/t10",
        {
            "instance": NotifyViber,
            "requests_response_text": VIBER_GOOD_RESPONSE,
            "privacy_url": "viber://****/t10",
        },
    ),
    (
        "viber://token/targetZ",
        {
            "instance": NotifyViber,
            # throw a bizarre code forcing us to fail to look it up
            "requests_response_text": VIBER_GOOD_RESPONSE,
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        "viber://token/targetZ",
        {
            "instance": NotifyViber,
            # force a failure
            "response": False,
            "requests_response_text": VIBER_BAD_RESPONSE,
            "requests_response_code": requests.codes.internal_server_error,
        },
    ),
    (
        "viber://token/targetY",
        {
            "instance": NotifyViber,
            # Throws a series of i/o exceptions with this flag
            # is set and tests that we gracefully handle them
            "requests_response_text": VIBER_GOOD_RESPONSE,
            "test_requests_exceptions": True,
        },
    ),
)


def test_plugin_viber_urls():
    """Verify URL parsing, privacy, and basic validation."""
    AppriseURLTester(tests=apprise_url_tests).run_all()


def test_plugin_viber_http_error_and_exception(mocker):
    """Verify HTTP error and requests exception paths."""
    post = mocker.patch("requests.post")

    post.return_value.status_code = 400
    a = Apprise()
    assert a.add("viber://token/target2") is True
    assert a.notify("test") is False

    post.side_effect = requests.RequestException("boom")
    a = Apprise()
    assert a.add("viber://token/target2") is True
    assert a.notify("test") is False
